#include "glwidget.h"

#include <QMouseEvent>

#include <GL/gl.h>
#include <GL/glu.h>

#include <cmath>
#include <cstdlib>
#include <ctime>

#include "superviewbox.h"

#include "ColourBasis.h"
#include "Slicer.h"

GLWidget::GLWidget(QWidget* parent)
	:QGLWidget(parent),mTimer(NULL),mMutationRate(.5),mClickMode(Random)
{
	setMode(Random);
	std::srand(std::time(NULL));

	for(int i=0;i<20;i++)
		mThings.push_back(SuperFormula::random());
}

GLWidget::~GLWidget()
{
	if (mTimer)
		mTimer->stop();
	delete mTimer;

	for(std::list<SuperFormula*>::iterator it = mThings.begin();it!=mThings.end();it++)
		delete (*it);

	makeCurrent();
}

QSize GLWidget::minimumSizeHint() const
{
	return QSize(200,200);
}

QSize GLWidget::sizeHint() const
{
	return QSize(800,800);
}

void GLWidget::initializeGL()
{
	glClearColor(.5,.1,.3,.8);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

 	//glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST);
 	//glEnable(GL_POLYGON_SMOOTH);
// glDisable(GL_DEPTH_TEST);

  // glCullFace(GL_BACK);
	// glEnable(GL_CULL_FACE);
	//

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// setup background texture
	Noise* noise = new Noise(16);
	double cols[] = {0,0.5,0.2,0.4,.1,
									.4,0.5,.2,.4,.1,
									.95,.8,.6,.2,.1,
									1,0.5,0.2,0.4,.1};
	ColourMap* cm = new ColourMap;
	cm->setBasis(noise,0);
	cm->getSpline()->set(cols,4);

	GLuint texture = SimpleSlicer::slice(*cm,512,512,16);
	glBindTexture(GL_TEXTURE_2D,texture);
}

void GLWidget::mutationRate(int k)
{
	mMutationRate = (double)k / 100;
}

void GLWidget::paintGL()
{
	glMatrixMode(GL_TEXTURE);
	glRotatef(0.01,0,0,1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glEnable(GL_TEXTURE_2D);
	glColor4f(1,1,1,.1);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex2f(-.5,+.5);
		glTexCoord2f(0,1);
		glVertex2f(-.5,-.5);
		glTexCoord2f(1,1);
		glVertex2f(+.5,-.5);
		glTexCoord2f(1,0);
		glVertex2f(+.5,+.5);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glScalef(0.1,0.1,0.1);

	// draw things

	for(std::list<SuperFormula*>::iterator it = mThings.begin();it!=mThings.end();it++)
		(*it)->draw();

	// draw last click
	/*
	glTranslatef(lastClick[0],lastClick[1],0);
	glColor4f(1,0,0,.5);
	glBegin(GL_QUADS);
		glVertex2f(-.5,-.5);
		glVertex2f(-.5,+.5);
		glVertex2f(+.5,+.5);
		glVertex2f(+.5,-.5);
	glEnd();
	*/

	/* view_all

	int numThings = mThings.size();
	int rows = std::ceil(std::sqrt((double)numThings));
	int cols = numThings / rows;
	
	int maxD = std::max(rows,cols);

	double scale = .8/maxD;
	
	glScalef(scale,scale,scale);
	glTranslatef(-cols/2+.5,-rows/2+.5,0);

	int row = 0, col = 0;
	for(std::list<SuperFormula*>::iterator it = mThings.begin();it!=mThings.end();it++)
	{
		if (col >= cols) 
		{
			col = 0;
			row++;
			glPopMatrix();
			glTranslatef(0,1,0);
		}
	
		if (col==0)
		{
			glPushMatrix();
		}
		(*it)->draw();
		glTranslatef(1,0,0);
		col++;
	}

	if (col>0) glPopMatrix();
	*/
}

void GLWidget::resizeGL(int width, int height)
{
	int size = std::min(width,height);
	glViewport((width-size)/2,(height-size)/2,size,size);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(-.5,+.5,+.5,-.5);
	glMatrixMode(GL_MODELVIEW);
}

void GLWidget::update()
{
	for(std::list<SuperFormula*>::iterator it = mThings.begin();it!=mThings.end();it++)
		(*it)->update(0.005);
	repaint();

}
void GLWidget::timer()
{
	if (mTimer==NULL)
	{
		mTimer = new QTimer(this);
		connect(mTimer,SIGNAL(timeout()),this,SLOT(update()));
		mTimer->start(5);
		return;
	}
	else
	{
		if (mTimer->isActive())
			mTimer->stop();
		else
			mTimer->start(5);
	}
}

void GLWidget::mousePressEvent(QMouseEvent* e)
{
	// map widget coords to world coords
	int size = std::min(width(),height());
	if (e->x() < (width()-size)/2 or e->x() > size+(width()-size)/2)
		return;
	if (e->y() < (height()-size)/2 or e->y() > size+(height()-size)/2)
		return;

	// else
	
	double x = 10*(1.0*(e->x() - (width()-size)/2) / size -.5);
	double y = 10*(1.0*(e->y() - (height()-size)/2) / size -.5);

	lastClick[0] = x;
	lastClick[1] = y;

	if (e->button()==Qt::RightButton)
		killThingAt(x,y);
	else if (e->button()==Qt::LeftButton)
	{
		if (mClickMode==Random)
			newThingAt(x,y);
		else if (mClickMode==Mutate)
			childrenAt(x,y,1);
		else if (mClickMode==ManyKids)
			childrenAt(x,y,5);
	}
	else observeAt(x,y);
}

void GLWidget::killThingAt(double x, double y)
{
	for(std::list<SuperFormula*>::iterator it = mThings.begin(); it!=mThings.end(); it++)
	{
		double dx = (*it)->x() - x;
		double dy = (*it)->y() - y;
		if (dx*dx + dy*dy < 1)
		{
			mThings.erase(it);
			break;
		}
	}
}

void GLWidget::newThingAt(double x, double y)
{
	SuperFormula* sf = SuperFormula::random();
	sf->x(x);
	sf->y(y);
	mThings.push_back(sf);
}

void GLWidget::observeAt(double x, double y)
{
	for(std::list<SuperFormula*>::iterator it = mThings.begin(); it!=mThings.end(); it++)
	{
		double dx = (*it)->x() - x;
		double dy = (*it)->y() - y;
		if (dx*dx + dy*dy < 1)
		{
			SuperViewBox* sf = new SuperViewBox(new SuperFormula(**it),this);
			break;
		}
	}
}

void GLWidget::childrenAt(double x, double y, int num)
{
	for(std::list<SuperFormula*>::iterator it = mThings.begin(); it!=mThings.end(); it++)
	{
		double dx = (*it)->x() - x;
		double dy = (*it)->y() - y;
		if (dx*dx + dy*dy < 1)
		{
			// make a baby!
			for(int i=0;i<num;i++)
				mThings.push_back((*it)->spawn(mMutationRate));
			break;
		}
	}
}


void GLWidget::killAll()
{
	for(std::list<SuperFormula*>::iterator it = mThings.begin(); it!=mThings.end(); it++)
	{
		delete *it;
	}
	mThings.clear();
}

void GLWidget::setMode(ClickMode c)
{
	mClickMode = c;

	if (c==Random)
	{
		setCursor(Qt::WhatsThisCursor);
	}
	else if (c==Mutate)
	{
		setCursor(Qt::UpArrowCursor);
	}
	else if (c==ManyKids)
	{
		setCursor(Qt::SizeAllCursor);
	}
}
