#include "super.h"

#include <cmath>
#include <cstdlib>
#include <GL/gl.h>

#include <QtGlobal>

const double BORDER = 6;

const int SuperFormula::sParamRanges[9][2] = 
{
	{10,100}, // a
	{10,100}, // b
	{10,200}, // m
	{1,1000}, // n1
	{1,1000}, // n2
	{1,1000}, // n3
	{-300,300}, // rotspeed
	{0,100}, // eyesize
	{0,100}, // border
};

const double SuperFormula::sParamMultiplier[9] =
{
	.1, // a
	.1, // b
	.1, // m
	1, // n1
	1, // n2
	1, // n3
	1, // rotspeed
	.05, // eyesize
	.01, // border
};

	SuperFormula::SuperFormula(double a, double b, double m, double n1, double n2, double n3, int divisions)
:mCached(false),mListId(0),mFactor(1),mRotation(0),mRotationSpeed(1),mBlinking(false),mNextBlink(1),mEyeSize(1)
	,mSpeed(1),mRecklessness(.1)
{
	if (divisions < 4) mDivs = 50 + 10*m;
	else mDivs = divisions;
	mParams[0] = a;
	mParams[1] = b;
	mParams[2] = m;
	mParams[3] = n1;
	mParams[4] = n2;
	mParams[5] = n3;

	mColours[0] = 1; mColours[1] = 1; mColours[2] = 0; mColours[3] = 1;// yellow outside
	mColours[4] = 1; mColours[5] = .3; mColours[6] = 0; mColours[7] = 1; // orange inside
	mColours[8] = 0; mColours[9] = .5; mColours[10] = .9;mColours[11] = 1; // blue iris

	mBorderThickness = 0.1;

	mPosition[0] = mPosition[1] = 0;
	mDirection[0] = 1; mDirection[1] = 0;
}

SuperFormula::SuperFormula(const SuperFormula& sf)
{
	mCached = false;
	mListId = 0;
	mFactor = 1;
	mRotation = sf.mRotation;
	mRotationSpeed = sf.mRotationSpeed;
	mBlinking = false;
	mNextBlink = 1;
	mEyeSize = sf.mEyeSize;
	mDivs = sf.mDivs;
	for(int i=0;i<6;i++)
		mParams[i] = sf.mParams[i];
	for(int i=0;i<12;i++)
		mColours[i] = sf.mColours[i];
	mBorderThickness = sf.mBorderThickness;
	mPosition[0] = sf.mPosition[0];
	mPosition[1] = sf.mPosition[1];
	mDirection[0] = sf.mDirection[0];
	mDirection[1] = sf.mDirection[1];
	mSpeed = sf.mSpeed;
	mRecklessness = sf.mRecklessness;
}



SuperFormula::~SuperFormula()
{
	if (mListId != 0)
		glDeleteLists(mListId,1);
}

void SuperFormula::setParam(int param, int val)
{
	getParam(param) = val*sParamMultiplier[param];

	if (param!=6 and param!=7)
		recache();
}

double& SuperFormula::getParam(int param)
{
/* order
	{1,10}, // a
	{1,10}, // b
	{1,10}, // m
	{1,10}, // n1
	{1,10}, // n2
	{1,10}, // n3
	{1,10}, // rotspeed
	{1,10}, // eyesize
	{1,10}, // border
	*/

	switch(param)
	{
		case 0: {return mParams[0]; break;}
		case 1: {return mParams[1]; break;}
		case 2: {return mParams[2]; break;}
		case 3: {return mParams[3]; break;}
		case 4: {return mParams[4]; break;}
		case 5: {return mParams[5]; break;}
		case 6: {return mRotationSpeed; break;}
		case 7: {return mEyeSize; break;}
		case 8: {return mBorderThickness; break;}
		default: return mBorderThickness; break;
	}
}

void SuperFormula::update(double dt)
{
	mRotation += mRotationSpeed*dt;

	mNextBlink -= dt;
	if (mNextBlink < 0)
	{
		mBlinking = not mBlinking;
		if (mBlinking)
			mNextBlink = 0.15;
		else
			mNextBlink = 2.0+(1.0*std::rand()/RAND_MAX);
	}

	mPosition[0] += mDirection[0]*mSpeed*dt;
	mPosition[1] += mDirection[1]*mSpeed*dt;

	mPosition[0] = toroidal_clamp(mPosition[0],-BORDER,BORDER);
	mPosition[1] = toroidal_clamp(mPosition[1],-BORDER,BORDER);

	rotate((mRecklessness*(2.0*std::rand()/RAND_MAX-1))*M_PI*dt);
}

void SuperFormula::buildList()
{
	if (mListId==0)
		mListId = glGenLists(1);

	glNewList(mListId,GL_COMPILE);

	const int& DIVISIONS = mDivs;
	const double dphi = 2.0*M_PI/DIVISIONS;
	const double md4 = mParams[2]/4.0;
	const double inva = 1.0/mParams[0];
	const double invb = 1.0/mParams[1];
	const double invn1 = 1.0/mParams[3];

	glBegin(GL_TRIANGLE_FAN);
	glColor4f(0,0,0,0);
	glVertex2f(0.,0.);
	glColor4f(0,0,0,1);
	double maxr = 0;
	for (int i=0;i<=DIVISIONS;i++)
	{
		double phi = dphi*i;
		double r = 1.0 / 
			std::pow(
				std::pow(
						std::abs(inva*std::cos(phi*md4)),
						mParams[4]
					) + 
					std::pow(
						std::abs(invb*std::sin(phi*md4)),
						mParams[5]
					),
				invn1);
	
		if (r > maxr) maxr = r;
		double x = r*std::cos(phi);
		double y = r*std::sin(phi);

		glVertex2f(x,y);
	}
	glEnd();

	glPushMatrix();
	double sc = 1-mBorderThickness;
	glScalef(sc,sc,sc);
	glBegin(GL_TRIANGLE_FAN);
	glColor4fv(mColours+4);
	glVertex2f(0.,0.);
	glColor4fv(mColours);
	for (int i=0;i<=DIVISIONS;i++)
	{
		double phi = dphi*i;
		double r = 1.0 / 
			std::pow(
				std::pow(
						std::abs(inva*std::cos(phi*md4)),
						mParams[4]
					) + 
					std::pow(
						std::abs(invb*std::sin(phi*md4)),
						mParams[5]
					),
				invn1);
	
		if (r > maxr) maxr = r;
		double x = r*std::cos(phi);
		double y = r*std::sin(phi);

		glVertex2f(x,y);
	}
	mFactor = 1.0/maxr;
	glEnd();
	glPopMatrix();
	glEndList();
}

void SuperFormula::draw()
{
	if (not mCached)
	{
		buildList();
	}

	glPushMatrix();
	glTranslatef(mPosition[0],mPosition[1],0);
	glScalef(mFactor,mFactor,mFactor);
	glRotatef(mRotation,0,0,1);
	glCallList(mListId);
	glRotatef(-mRotation,0,0,1);
	glScalef(mEyeSize,mEyeSize,mEyeSize);
	drawEye();
	glPopMatrix();
}

void SuperFormula::drawPreview()
{
	if (not mCached)
	{
		buildList();
	}

	glPushMatrix();
	glScalef(mFactor,mFactor,mFactor);
	glRotatef(mRotation,0,0,1);
	glCallList(mListId);
	glRotatef(-mRotation,0,0,1);
	glScalef(mEyeSize,mEyeSize,mEyeSize);
	drawEye();
	glPopMatrix();
}

void SuperFormula::drawEye()
{
	glPushMatrix();
	glScalef(.3,.3,.3);

	if (mBlinking)
	{
		glColor3f(0,0,0);
		glBegin(GL_LINES);
			glVertex2f(-1,0);
			glVertex2f(1,0);
		glEnd();
	}
	else
	{
		glColor3f(1,1,1);
		fillCircle();
		glScalef(.6,.6,.6);
		glColor4fv(mColours+8);
		fillCircle();
		glScalef(.5,.5,.5);
		glColor3f(0,0,0);
		fillCircle();
	}
	glPopMatrix();
}

void SuperFormula::fillCircle()
{	
	double dphi = M_PI / 30.0;
	glBegin(GL_TRIANGLE_FAN);
	glVertex2f(0,0);
	for(int i=0;i<=60;i++)
	{
		double phi = dphi*i;
		glVertex2f(std::cos(phi),std::sin(phi));
	}
	glEnd();
}

SuperFormula*
SuperFormula::random()
{
	// params range from 1..?
	const double pr[6] = {
		3,
		3,
		10,
		1000,
		1000,
		1000
	};
	
	// colours can be anything...
	
	// rotation speed range -10--10

	// border thickness, 0--.3
	
	// eye size .5--3
	
	// recklessness 0-.5
	
	// speed 0-10
	
	// position in [-BORDER,BORDER]
	
	// direction [0,1] rotated by a random amount
	
	SuperFormula* sf = new SuperFormula(randD(1,pr[0]),randD(1,pr[1]),randD(1,pr[2]),randD(1,pr[3]),randD(1,pr[4]),randD(1,pr[5]));
	for(int i=0;i<12;i++)
		sf->mColours[i] = randD(0,1);

	sf->mBorderThickness = randD(0,.3);
	sf->mRotationSpeed = randD(-10,10);
	sf->mEyeSize = randD(.5,3);
	sf->mRecklessness = randD(.9,10);
	sf->mSpeed = randD(0,2);
	sf->mPosition[0] = randD(-BORDER,BORDER);
	sf->mPosition[1] = randD(-BORDER,BORDER);
	sf->rotate(randD(0,2.0*M_PI));
	return sf;
}

double SuperFormula::randD(double from, double to)
{
	return (1.0*std::rand()/RAND_MAX)*(to-from) + from;
}

void SuperFormula::rotate(double angle)
{
	double x = mDirection[0];
	double y = mDirection[1];

	mDirection[0] = x*std::cos(angle) + y*std::sin(angle);
	mDirection[1] = -x*std::sin(angle) + y*std::cos(angle);
}

double SuperFormula::toroidal_clamp(double x, double leastx, double mostx)
{
	double diff = mostx-leastx;
	while (x < leastx) 
		x += diff;
	while (x > mostx)
		x -= diff;
	return x;
}

SuperFormula* SuperFormula::spawn(double mutationFactor)
{
	double low = 1-mutationFactor;
	double high = 1+mutationFactor;

	double p[6];
	for(int i=0;i<6;i++)
	{
		p[i] = mParams[i]*randD(low,high);
		p[i] = qMax(p[i],0.);
	}

	SuperFormula* sf = new SuperFormula(p[0],p[1],p[2],p[3],p[4],p[5]);
	
	for(int i=0;i<12;i++)
	{
		GLfloat& c = sf->mColours[i];
		c = mColours[i]*randD(low,high);
		c = qMax(0.f,qMin(c,1.f));
	}

	sf->mBorderThickness = mBorderThickness*randD(low,high);
	sf->mRotationSpeed = mRotationSpeed*randD(low,high);
	sf->mEyeSize = mEyeSize*randD(low,high);
	sf->mRecklessness = mRecklessness*randD(low,high);
	sf->mSpeed = mSpeed*randD(low,high);
	sf->mPosition[0] = mPosition[0] + randD(-1,1);
	sf->mPosition[1] = mPosition[1] + randD(-1,1);
	sf->rotate(randD(0,2.0*M_PI));
	return sf;
}

std::string SuperFormula::getParamName(int param)
{
	switch (param)
	{
		case 0: return "a";
		case 1: return "b";
		case 2: return "m";
		case 3: return "n1";
		case 4: return "n2";
		case 5: return "n3";
		case 6: return "rotspeed";
		case 7: return "eye";
		default: return "border";
	};
}
